#include <memory>
#include <csignal>
#include <spdlog/sinks/basic_file_sink.h>
#include <boost/asio/co_spawn.hpp>
#include <boost/asio/signal_set.hpp>
#include "server/server.hxx"
#include "util/botan.hxx"
#include "index_html.h"
#include "favicon_svg.h"

using tcp = boost::asio::ip::tcp;

std::function<void()> g_stop_service;
std::shared_ptr<spdlog::logger> err_handler::logger{};

Server server{};
std::string index_etag{};

net::awaitable<void> do_session(tcp_stream stream) {
    tcp_buffer buffer;
    beast::error_code ec;
    stream.expires_after(std::chrono::seconds(30));

    try {
        while (true) {
            http::request_parser<http::empty_body> req_base_parser;
            req_base_parser.body_limit(boost::none);
            co_await http::async_read_header(stream, buffer, req_base_parser);

            auto &req_base = req_base_parser.get();
            bool keep_alive = req_base.keep_alive();

            auto addr = stream.socket().remote_endpoint().address().to_string();
            if (!req_base["x-forwarded-for"].empty())
                addr = std::string{req_base["x-forwarded-for"]} + "' via '" + addr;
            spdlog::info("Request: '{} {}' from '{}'", req_base.method_string(), req_base.target(), addr);

            if (req_base.method() == http::verb::post && req_base.target() == "/mrpc") {
                req_base_parser.body_limit(1024*1024*8); // 8 MB
                http::request_parser<http::string_body> req{std::move(req_base_parser)};
                co_await http::async_read(stream, buffer, req);
                co_await server.msg_handler(stream, req.get());
            } else if (req_base.method() == http::verb::get && req_base.target() == "/") {
                co_await http::async_read(stream, buffer, req_base_parser);
                if (req_base["If-None-Match"] == index_etag) {
                    auto res = create_response<http::status::not_modified, http::empty_body, http::empty_body>(req_base);
                    res.set(http::field::cache_control, "no-cache");
                    res.set(http::field::etag, index_etag);
                    co_await http::async_write(stream, res, net::use_awaitable);
                } else {
                    auto res = create_response<http::status::ok, http::buffer_body, http::empty_body>(req_base);
                    res.set(http::field::content_type, "text/html");
                    res.content_length(index_html_len);
                    res.set(http::field::cache_control, "no-cache");
                    res.set(http::field::etag, index_etag);
                    res.body().data = index_html;
                    res.body().size = index_html_len;
                    res.body().more = false;
                    co_await http::async_write(stream, res, net::use_awaitable);
                }
            } else if (req_base.method() == http::verb::post && req_base.target() == "/download") {
                req_base_parser.body_limit(1024*1024*8);
                http::request_parser<http::string_body> req{std::move(req_base_parser)};
                co_await http::async_read(stream, buffer, req);
                co_await server.download(stream, req.get());
            } else if (req_base.method() == http::verb::post && req_base.target() == "/upload") {
                http::request_parser<http::file_body> req_parser{std::move(req_base_parser)};
                co_await server.upload(stream, buffer, req_parser);
                keep_alive = false;
            } else if (req_base.method() == http::verb::post && req_base.target() == "/download_multi") {
                req_base_parser.body_limit(1024*1024*8);
                http::request_parser<http::string_body> req{std::move(req_base_parser)};
                co_await http::async_read(stream, buffer, req);
                co_await server.download_multi(stream, req.get());
            } else if (req_base.method() == http::verb::get && req_base.target() == "/favicon.svg") {
                co_await http::async_read(stream, buffer, req_base_parser);
                auto res = create_response<http::status::ok, http::buffer_body, http::empty_body>(req_base);
                res.set(http::field::content_type, "image/svg+xml");
                res.content_length(favicon_svg_len);
                res.body().data = favicon_svg;
                res.body().size = favicon_svg_len;
                res.body().more = false;
                co_await http::async_write(stream, res, net::use_awaitable);
            } else {
                http::response<http::string_body> res{http::status::not_found, req_base.version()};
                res.set(http::field::content_type, "text/plain");
                res.keep_alive(false);
                keep_alive = false;
                res.body() = "404 not found";
                res.prepare_payload();
                co_await http::async_write(stream, res, net::use_awaitable);
            }

            if (!keep_alive)
                break;
        }
    } catch (beast::system_error &se) {
        if (se.code() != http::error::end_of_stream)
            throw;
    }

    ec = stream.socket().shutdown(tcp::socket::shutdown_send, ec);
}

net::awaitable<void> do_listen(tcp::endpoint endpoint) {
    auto acceptor = net::use_awaitable.as_default_on(tcp::acceptor(co_await net::this_coro::executor));

    acceptor.open(endpoint.protocol());
    acceptor.set_option(net::socket_base::reuse_address(true));
    acceptor.bind(endpoint);
    acceptor.listen(net::socket_base::max_listen_connections);

    while (true) {
        net::co_spawn(
            acceptor.get_executor(),
            do_session(tcp_stream(co_await acceptor.async_accept())),
            err_handler{}
        );
    }
}

int main() {
    // TODO add current timestamp to log name
    auto file_sink = std::make_shared<spdlog::sinks::basic_file_sink_mt>("log.txt");
    spdlog::default_logger()->sinks().push_back(file_sink);
    spdlog::set_level(spdlog::level::trace);

    {
        auto md5_hash = Botan::HashFunction::create_or_throw("MD5");
        md5_hash->update(index_html, index_html_len);
        index_etag = "\"" + Botan::hex_encode(md5_hash->final()) + "\"";
    }

    net::io_context ioc{server.config.threads};
    net::co_spawn(ioc, do_listen(tcp::endpoint{net::ip::address_v4::any(), server.config.server_port}), err_handler{});

    net::signal_set signals{ioc, SIGINT, SIGTERM};
    signals.async_wait([&ioc](const beast::error_code&, int){ ioc.stop(); });
    g_stop_service = [&ioc] { ioc.stop(); };

    spdlog::info("Listening on :{}", server.config.server_port);

    std::vector<std::thread> v;
    for (std::uint16_t i = server.config.threads-1; i > 0; --i)
        v.emplace_back([&ioc] { ioc.run(); });
    ioc.run();

    for (auto &t : v)
        t.join();

    return 0;
}
