#include <fstream>
#include <spdlog/spdlog.h>
#include <stb_image.h>
#include <stb_image_resize2.h>
#include <stb_image_write.h>
#include "server_internal.hxx"

static constexpr std::size_t chunk_size = 1024*1024, max_image_size = 1024*1024*50, preview_size=480;
static const std::set<std::string> image_extension = {".png", ".jpg", ".jpeg", ".tga", ".bmp", ".psd", ".gif", ".jfif", ".pjpeg", ".pjp"};

struct UploadInfo {
    Server *server;
    std::shared_lock<std::shared_mutex> node_lock;
    std::size_t to_read;
    std::filesystem::path path;
    std::ofstream file;
    std::shared_ptr<Node> node;
};

void make_preview(std::filesystem::path &path, const std::shared_ptr<Node> &node) {
    int x, y, channels;
    auto img = std::unique_ptr<stbi_uc, decltype(&free)>
        {stbi_load(path.c_str(), &x, &y, &channels, 0), &free};
    if (!img)
        return;

    float x_ration = (float)preview_size / (float)x, y_ration = (float)preview_size / (float)y;
    float ratio = std::min(x_ration, y_ration);
    int new_x = (int)((float)(x)*ratio), new_y = (int)((float)(y)*ratio);
    stbir_pixel_layout layout;
    switch (channels) {
        case 1: layout = STBIR_1CHANNEL; break;
        case 2: layout = STBIR_2CHANNEL; break;
        case 3: layout = STBIR_RGB; break;
        case 4: layout = STBIR_RGBA; break;
        default: return;
    }

    auto rimg = std::unique_ptr<unsigned char, decltype(&free)>
        {stbir_resize_uint8_linear(img.get(), x, y, 0, nullptr, new_x, new_y, 0, layout), &free};
    if (!rimg)
        return;

    auto png_path = path.replace_extension("png");
    if (!stbi_write_png(png_path.c_str(), new_x, new_y, channels, rimg.get(), 0))
        return;

    node->preview = true;
}

net::awaitable<void> Server::upload(tcp_stream &s, tcp_buffer &buf, http::request_parser<http::file_body> &req) {
    auto &body = req.get();

    if (body["X-Node"].empty()) { co_await send_error(s, body, "Missing node"); co_return; }
    if (body["X-Token"].empty()) { co_await send_error(s, body, "Missing token"); co_return; }

    if (body[http::field::transfer_encoding] == "chunked") {
        spdlog::critical("Encountered a chunked upload!");
        co_await send_error(s, body, "Sorry but your browser is not supported yet"); co_return;
    }

    auto node_str = body["X-Node"];
    std::uint64_t node_id = 0;
    std::from_chars(node_str.data(), node_str.data() + node_str.size(), node_id);
    std::string token = body["X-Token"];

    check_user() { co_await send_error(s, body, "Invalid user"); co_return; }

    auto node = get_node(user, node_id);
    if (!node) { co_await send_error(s, body, "Invalid node"); co_return; }
    if (!node->file) { co_await send_error(s, body, "Can't upload to a directory"); co_return; }

    auto path = user->user_dir / std::to_string(node->id);
    if (node->preview) {
        node->preview = false;
        std::filesystem::remove(path.replace_extension("png"));
    }

    std::exception_ptr ex_ptr;
    beast::error_code ec;
    body.body().open(path.c_str(), beast::file_mode::write, ec);
    try {
        co_await http::async_read(s, buf, req);
    } catch (beast::system_error &se) {
        ex_ptr = make_exception_ptr(se);
    }
    body.body().close();

    if (ex_ptr)
        std::filesystem::resize_file(path, 0);

    std::size_t real_size = std::filesystem::file_size(path);
    node->size = real_size;
    auto ext = std::filesystem::path{node->name}.extension().string();
    if (real_size > 0 && real_size < max_image_size && image_extension.contains(ext))
        make_preview(path, node);

    save();

    auto res = create_response<http::status::ok, http::empty_body, http::file_body>(body);
    res.keep_alive(false);
    co_await http::async_write(s, res, net::use_awaitable);

    if (ex_ptr)
        rethrow_exception(ex_ptr);
}
